from __future__ import absolute_import, division, print_function, unicode_literals

import os, sys, hashlib, base64, zlib
from ctypes import cast, memmove, POINTER, c_void_p
from .structs import EVPobject

opaque_repr = False

class ResumableHasher(object):
    name = None
    _algorithms_guaranteed = getattr(hashlib,
                                     "algorithms_guaranteed",
                                     ["md5", "sha1", "sha224", "sha256", "sha384", "sha512"])

    def __init__(self, name=None, data=None, state=None):
        if state is not None:
            if not self.name:
                raise Exception('Parameter "name" is required')
            self.__setstate__(state=dict(name=name, md_data=zlib.decompress(base64.b64decode(state))))
            if data is not None:
                self.update(data)
            return
        if self.name is not None:
            data = name
        else:
            self.name = name
        if not self.name:
            raise Exception('Parameter "name" is required')
        hasher_args = [] if data is None else [data]
        self._hasher = self._get_hashlib_hasher(self.name)(*hasher_args)

    def _get_hashlib_hasher(self, name):
        if name.startswith("blake2"):
            raise Exception("blake2 algorithms are not OpenSSL-based and not supported by rehash")
        if name.startswith("sha3"):
            raise Exception("sha3 algorithms are not supported by rehash")
        if name.startswith("shake"):
            raise Exception("shake algorithms are not supported by rehash")
        if name in self._algorithms_guaranteed:
            return getattr(hashlib, name)
        else:
            return hashlib.new(name)

    def _get_evp_md_ctx(self):
        c_evp_obj = cast(c_void_p(id(self._hasher)), POINTER(EVPobject))
        if hasattr(c_evp_obj.contents.ctx, "contents"):
            return c_evp_obj.contents.ctx.contents
        else:
            return c_evp_obj.contents.ctx

    def __getstate__(self):
        ctx = self._get_evp_md_ctx()
        ctx_size = ctx.digest.contents.ctx_size
        hasher_state = ctx.md_data[:ctx_size]
        return dict(name=self.name, md_data=hasher_state)

    def __setstate__(self, state):
        self.name = state["name"]
        self._hasher = self._get_hashlib_hasher(self.name)()
        ctx = self._get_evp_md_ctx()
        ctx_size = ctx.digest.contents.ctx_size
        memmove(ctx.md_data, state["md_data"], ctx_size)

    def __getattr__(self, a):
        return getattr(self._hasher, a)

    def __repr__(self):
        if opaque_repr:
            return "{}.{}()".format(self.__module__, self.__class__.__name__)
        else:
            md_data = base64.b64encode(zlib.compress(self.__getstate__()["md_data"])).decode()
            return "{}.{}(state='{}')".format(self.__module__, self.name, md_data)


new = ResumableHasher

def _initialize():
    module = sys.modules[__name__]
    for name in ResumableHasher._algorithms_guaranteed:
        if name.startswith("blake2") or name.startswith("sha3") or name.startswith("shake"):
            continue
        setattr(module, name, type(name, (ResumableHasher,), dict(name=name)))


_initialize()



from __future__ import absolute_import, division, print_function, unicode_literals

from ssl import OPENSSL_VERSION
from sys import version_info as PYTHON_VERSION
from ctypes import c_void_p, POINTER, Structure, c_int, c_ulong, c_char, c_size_t, c_ssize_t, py_object

PyObject_HEAD = [
    ('ob_refcnt', c_size_t),
    ('ob_type', c_void_p)
]

# OpenSSL 1.0.2 and earlier:
# https://github.com/openssl/openssl/blob/OpenSSL_1_0_2-stable/crypto/evp/evp.h#L159-L181
# OpenSSL 1.1.0 and later:
# https://github.com/openssl/openssl/blob/master/crypto/include/internal/evp_int.h#L99-L113
class EVP_MD(Structure):
    _fields_ = [
        ('type', c_int),
        ('pkey_type', c_int),
        ('md_size', c_int),
        ('flags', c_ulong),
        ('init', c_void_p),
        ('update', c_void_p),
        ('final', c_void_p),
        ('copy', c_void_p),
        ('cleanup', c_void_p),
    ]
    if OPENSSL_VERSION < "OpenSSL 1.1.0":
        _fields_ += [
            ('sign', c_void_p),
            ('verify', c_void_p),
            ('required_pkey_type', c_int * 5),
        ]
    _fields_ += [
        ('block_size', c_int),
        ('ctx_size', c_int),
    ]

# https://github.com/openssl/openssl/blob/master/crypto/evp/evp_locl.h#L12-L22
class EVP_MD_CTX(Structure):
    _fields_ = [
        ('digest', POINTER(EVP_MD)),
        ('engine', c_void_p),
        ('flags', c_ulong),
        ('md_data', POINTER(c_char)),
    ]

# Python 3.8+: https://github.com/python/cpython/blob/master/Modules/_hashopenssl.c#L58-L64
# Python 3.5.3 - 3.7: https://github.com/python/cpython/blob/3.7/Modules/_hashopenssl.c#L52-L59
# Python 3 - 3.5.2: https://github.com/python/cpython/blob/3.4/Modules/_hashopenssl.c#L39-L46
# Python 2.7.13+: https://github.com/python/cpython/blob/2.7/Modules/_hashopenssl.c#L71-L78
class EVPobject(Structure):
    _fields_ = PyObject_HEAD + (
        [("name", POINTER(py_object))] if PYTHON_VERSION < (3, 8) else []
    ) + [
        ("ctx", EVP_MD_CTX if (3, 0) < PYTHON_VERSION < (3, 5, 3) or PYTHON_VERSION < (2, 7, 13) else POINTER(EVP_MD_CTX))
    ]
